/* EmuTO8 -- sound support
   Copyright (C) 1998 Jerome Thoen */

#include <stdio.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>

#include "config.h"
#include "audio.h"

#define FRAG_EXP 9
#define BUFSIZE (1L<<FRAG_EXP)

#define DEVNAME "/dev/dsp"

static int audio_fd;
static unsigned char audio_buffer[BUFSIZE];
static unsigned char *audio_bufptr;

__inline__ void audio_put_byte_and_sync(unsigned char b)
{
  *audio_bufptr++ = b;
  if (audio_bufptr >= audio_buffer + BUFSIZE) {
    audio_buf_info info;

    ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info);
    if (info.bytes <= BUFSIZE - BUFSIZE/100) {
      struct timeval tv;

      tv.tv_sec = 0;
      tv.tv_usec = 90 * (BUFSIZE - info.bytes);
      select(0, NULL, NULL, NULL, &tv);
    }
    write(audio_fd, audio_buffer, BUFSIZE);
    audio_bufptr = audio_buffer;
  }
}

int audio_init(void)
{
  int format = AFMT_U8, stereo = 0, speed = 11025;
  int arg = 0x00040000 + FRAG_EXP;

  audio_fd = open(DEVNAME, O_WRONLY);

  if (audio_fd == -1) {
    perror(DEVNAME);
    return 0;
  }

  if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &arg) == -1) {
    perror("SNCCTL_DSP_SETFRAGMENT");
    close(audio_fd);
    return 0;
  }

  if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
    perror("SNDCTL_DSP_SETFMT");
    close(audio_fd);
    return 0;
  }

  if (format != AFMT_U8) {
    printf("Le systeme ne supporte pas le son 8 bits\n");
    close(audio_fd);
    return 0;
  }

  if (ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo) == -1) {
    perror("SNDCTL_DSP_STEREO");
    close(audio_fd);
    return 0;
  }

  if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &speed) == -1) {
    perror("SNDCTL_DSP_SPEED");
    close(audio_fd);
    return 0;
  }

  printf("Audio using 8 bits mono 11025 Hz %ld bytes buffer\n", BUFSIZE);

  audio_bufptr = audio_buffer;

  return 1;
}

void audio_quit(void)
{
  close(audio_fd);
}

