/*	  

	VERSION     : 2.0
        MODULE      : disk.c
	Cr par    : Gilles FETIS
        Modifi par : Alexandre PUKALL Mai 1998

        Gestion du format SAP 1.0 lecture et ecriture disquette
        
*/

/*
 * 19/08/1998 Jerome Thoen
 * Modification pour utilisation dans emuto8
 */

#include "config.h"

#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "emuto8.h"
#include "memory.h"
#include "zfile.h"

#define NBSIDE 2
#define NBTRACK 80
#define NBSECT 16
#define SECTSIZE 256

unsigned short ax,bx,cx,dx,si,tmp,x1a2,x1a0[5],res1,i2,inter,cfc,cfd,compte,fois;
unsigned char cle[11],entree;
unsigned char bufferc[256];
int DRV,TRK,SEC,BUF;

int SAPOK=0;
char SAPfname[2][1024];
int erreur=0;
int ret=0;

/*
*/

typedef struct status_t
{
  unsigned char format;
  unsigned char protection;
  unsigned char piste;
  unsigned char secteur;
  unsigned char crc1sect;
  unsigned char crc2sect;
} status_t;

unsigned char disk[NBSIDE][NBTRACK][NBSECT][SECTSIZE];
status_t    disk_status[NBSIDE][NBTRACK][NBSECT];

short crcpuk_temp;


static const short table[]={

   0x0000, 0x1081, 0x2102, 0x3183,
   0x4204, 0x5285, 0x6306, 0x7387,
   0x8408, 0x9489, 0xa50a, 0xb58b,
   0xc60c, 0xd68d, 0xe70e, 0xf78f

 };

void crc_pukall(short c)
{
    register short i,index;

    index = (crcpuk_temp ^ c) & 0xf;
    crcpuk_temp = ((crcpuk_temp>>4) & 0xfff) ^ table[index];

    c >>= 4;

    index = (crcpuk_temp ^ c) & 0xf;
    crcpuk_temp = ((crcpuk_temp>>4) & 0xfff) ^ table[index];
 }

int init_pc1(unsigned char drive1,unsigned char track1,unsigned char sect1)
{
   unsigned int charge;

   cle[0]=0x42 ^ SEC-1;
   cle[1]=0x2A ^ TRK;
   cle[2]=0x94 ^ disk_status[drive1][track1][sect1].crc1sect;
   cle[3]=0xC4 ^ disk_status[drive1][track1][sect1].crc2sect;
   cle[4]=0x09;
   cle[5]=0x29;
   cle[6]=0x13;
   cle[7]=0x15;
   cle[8]=0xA9;
   cle[9]=0x22;

   decode_pc1();

   if (get_memb(0x604A)<=0x3F )
    {
     entree=crcpuk_temp>>8;
     if ( entree != disk_status[drive1][track1][sect1].crc1sect )
     {
      return(1);
     }

     entree=crcpuk_temp&255;
     if ( entree != disk_status[drive1][track1][sect1].crc2sect )
     {
      return(1);
     }
     return(0);
   }

   for (fois=0;fois<256;fois++)
    {
     bufferc[fois]=disk[DRV][TRK][SEC-1][fois];
    }

   charge=0;
   charge=(get_memb(0x604A)<<8);

   cle[0]=get_memb(charge) ^ SEC-1;
   cle[1]=get_memb(charge+1) ^ TRK;
   cle[2]=get_memb(charge+2) ^ disk_status[drive1][track1][sect1].crc1sect;
   cle[3]=get_memb(charge+3) ^ disk_status[drive1][track1][sect1].crc2sect;
   cle[4]=get_memb(charge+4);
   cle[5]=get_memb(charge+5);
   cle[6]=get_memb(charge+6);
   cle[7]=get_memb(charge+7);
   cle[8]=get_memb(charge+8);
   cle[9]=get_memb(charge+9);

    decode_pc1();

     entree=crcpuk_temp>>8;
     if ( entree != disk_status[drive1][track1][sect1].crc1sect )
     {
      return(1);
     }

     entree=crcpuk_temp&255;
     if ( entree != disk_status[drive1][track1][sect1].crc2sect )
     {
      return(1);
     }
     return(0);

}

int verify_sap_ecri(unsigned char drive,unsigned char track,unsigned char sect)
{
  if ( (disk_status[drive][track][sect].protection==1) || (disk_status[drive][track][sect].protection==3) )
  {
   erreur=1;
   return(1);
  }

  if (disk_status[drive][track][sect].format==5)
  {
   erreur=4;
   return(1);
  }

    crcpuk_temp = 0xffff;

    crc_pukall(disk_status[drive][track][sect].format);
    crc_pukall(disk_status[drive][track][sect].protection);
    crc_pukall(disk_status[drive][track][sect].piste);
    crc_pukall(disk_status[drive][track][sect].secteur);

    for (fois=0;fois<=255;fois++)
     {
       crc_pukall( get_memb((BUF+fois)%0xFFFF) );
     }

     entree=crcpuk_temp>>8;
     disk_status[drive][track][sect].crc1sect=entree;

     entree=crcpuk_temp&255;
     disk_status[drive][track][sect].crc2sect=entree;

     erreur=0;
     return(0);

}

int verify_sap_lect(unsigned char drive,unsigned char track,unsigned char sect)
{

     if ( disk_status[drive][track][sect].format==5)
     {
       for (fois=0;fois<256;fois++)
        {
           bufferc[fois]=disk[drive][track][sect][fois];
        }
       erreur=init_pc1(drive,track,sect);
       if (erreur==0)
        {
          for (fois=0;fois<256;fois++)
          {
           set_memb(((BUF+fois)%0xFFFF),bufferc[fois]);
          }

          erreur=0;
          return(0);
        }
       else
        {
          erreur=4;
          return(1);
        }
     }

     erreur=0;
     return(0);

}


int verify_sap(unsigned char drive,unsigned char track,unsigned char sect)
{
    crcpuk_temp = 0xffff;

    crc_pukall(disk_status[drive][track][sect].format);
    crc_pukall(disk_status[drive][track][sect].protection);
    crc_pukall(disk_status[drive][track][sect].piste);
    crc_pukall(disk_status[drive][track][sect].secteur);

    if ( (track != disk_status[drive][track][sect].piste) || ((sect+1) != disk_status[drive][track][sect].secteur) )
    {
      erreur=4;
      return(1);
    }

    for (fois=0;fois<=255;fois++)
     {
       crc_pukall(disk[drive][track][sect][fois]);
     }

     entree=crcpuk_temp>>8;
     if ( entree != disk_status[drive][track][sect].crc1sect )
     {
      erreur=8;
      return(1);
     }

     entree=crcpuk_temp&255;
     if ( entree != disk_status[drive][track][sect].crc2sect )
     {
      erreur=8;
      return(1);
     }

     if ( disk_status[drive][track][sect].format==4)
     {
      erreur=4;
      return(1);
     }

     erreur=0;
     return(0);
}

FILE *OpenSAP(char *name)
{
  FILE *fp;
  char temp[256];

  if (fp = fopen(name, "rb"))
    return fp;
  strncpy(temp, name, 247);
  strcat(temp, ".sap");
  if (fp = fopen(temp, "rb"))
    return fp;
  strcpy(temp, "sap/");
  strncat(temp, name, 247);
  if (fp = fopen(temp, "rb"))
    return fp;
  strcat(temp, ".sap");
  if (fp = fopen(temp, "rb"))
    return fp;
  return NULL;
}

#define SAP_HEADER_SIZE 66
#define SAP_SECT_SIZE 262

int LoadSAP(char * name,int side)
{
   FILE * f;
   unsigned char buff[4096];
   int i,j,k;

   
   f = OpenSAP(name);

   if (f == NULL)
     return 0;

   /* mmorise le nom de l'archive pour uilisation ultrieure */
   strcpy(SAPfname[side],name);

   fread(buff,sizeof(char),(size_t)SAP_HEADER_SIZE,f);

   for (i=0;i<80;i++)
   for (j=0;j<16;j++)
   {
       int SEC,TRK;
       SEC=j;
       TRK=i;
       fread(buff,sizeof(char),(size_t)SAP_SECT_SIZE,f);
       disk_status[side][TRK][SEC].format=buff[0];
       disk_status[side][TRK][SEC].protection=buff[1];
       disk_status[side][TRK][SEC].piste=buff[2];
       disk_status[side][TRK][SEC].secteur=buff[3];

       for (k=0;k<SECTSIZE;k++)
          disk[side][TRK][SEC][k]=(buff[4+k]^0xB3);
       disk_status[side][TRK][SEC].crc1sect=buff[4+k];
       disk_status[side][TRK][SEC].crc2sect=buff[4+k+1];
   }
   fclose (f);

   /* les extensions SAP sont actives sur ce disk */
   SAPOK=1;
   return 1;
}

void InitDiskCtrl(void)
{
  setcc(getcc() & 0xfe);
  set_memb(0x604E,'D');
  return;
}

void ReadSector(void)
{
   int i,j;

   DRV=get_memb(0x6049);
   TRK=(get_memb(0x604A)<<8)|get_memb(0x604B);
   SEC=get_memb(0x604C);

   BUF=(get_memb(0x604F)<<8)|get_memb(0x6050);

   if ((DRV>1) || (TRK>=NBTRACK) || (SEC>NBSECT))
   {
      setcc(getcc() | 0x01);
      set_memb(0x604E,0x10);
      return;
   }

   if (SAPOK==0)
      for (i=0;i<256;i++) set_memb(((BUF+i) & 0xFFFF),disk[DRV][TRK][SEC-1][i]);
   else
   {
      ret=verify_sap(DRV,TRK,SEC-1);
      if (ret==0)
       {
         ret=verify_sap_lect(DRV,TRK,SEC-1);
         if (ret==0)
            for (i=0;i<256;i++) set_memb(((BUF+i) & 0xFFFF),disk[DRV][TRK][SEC-1][i]);
       }
   }

   if (ret==0)
   {
      set_memb(0x604E,0x00);
      setcc(getcc() & 0xfe);
   }
   else
   {
      set_memb(0x604E,erreur);
      setcc(getcc() | 0x01);
   }
}

void WriteSector(void)
{
   int i,j;

   DRV=get_memb(0x6049);
   TRK=(get_memb(0x604A)<<8)|get_memb(0x604B);
   SEC=get_memb(0x604C);

   BUF=(get_memb(0x604F)<<8)|get_memb(0x6050);

   if ((DRV>1) || (TRK>=NBTRACK) || (SEC>NBSECT))
   {
      setcc(getcc() | 0x01);
      set_memb(0x604E,0x10);
      return;
   }

   if (SAPOK==0)
      for (i=0;i<256;i++) disk[DRV][TRK][SEC-1][i]=get_memb((BUF+i)%0xFFFF);
   else
   {
      ret=verify_sap(DRV,TRK,SEC-1);
      if (ret==0)
      {
         ret=verify_sap_ecri(DRV,TRK,SEC-1);
         if (ret==0)
         {
            int fSAP;
            for (i=0;i<256;i++) disk[DRV][TRK][SEC-1][i]=get_memb((BUF+i)%0xFFFF);
            fSAP=open(SAPfname[DRV],O_RDWR);
            if (fSAP>0)
            {
		int i;
		unsigned char temp[SECTSIZE+2];
               size_t posInSAP;
               posInSAP= SAP_HEADER_SIZE + (TRK*NBSECT+(SEC-1))*SAP_SECT_SIZE + 4;
               lseek(fSAP,posInSAP,SEEK_SET);
		for (i=0;i<SECTSIZE;i++) temp[i]=disk[DRV][TRK][SEC-1][i]^0xB3;
		temp[i]=disk_status[DRV][TRK][SEC-1].crc1sect;
		temp[i+1]=disk_status[DRV][TRK][SEC-1].crc2sect;
               write(fSAP,temp,SECTSIZE+2);
               close(fSAP);
            }
         }
      }
   }

   if (ret==0)
   {
      set_memb(0x604E,0x00);
      setcc(getcc() & 0xfe);
   }
   else
   {
      set_memb(0x604E,erreur);
      setcc(getcc() | 0x01);
   }
}

void code(void)
{
   dx=x1a2+i2;
   ax=x1a0[i2];
   cx=0x015a;
   bx=0x4e35;

   tmp=ax;
   ax=si;
   si=tmp;

   tmp=ax;
   ax=dx;
   dx=tmp;

   if (ax!=0)
    {
      ax=ax*bx;
    }

   tmp=ax;
   ax=cx;
   cx=tmp;

   if (ax!=0)
    {
      ax=ax*si;
      cx=ax+cx;
    }

   tmp=ax;
   ax=si;
   si=tmp;
   ax=ax*bx;
   dx=cx+dx;

   ax=ax+1;

   x1a2=dx;
   x1a0[i2]=ax;

   res1=ax^dx;
   i2=i2+1;
 }

void assemble(void)
{

   x1a0[0]= ( cle[0]*256 )+ cle[1];
   code();
   inter=res1;

   x1a0[1]= x1a0[0] ^ ( (cle[2]*256) + cle[3] );
   code();
   inter=inter^res1;

   x1a0[2]= x1a0[1] ^ ( (cle[4]*256) + cle[5] );
   code();
   inter=inter^res1;

   x1a0[3]= x1a0[2] ^ ( (cle[6]*256) + cle[7] );
   code();
   inter=inter^res1;

   x1a0[4]= x1a0[3] ^ ( (cle[8]*256) + cle[9] );
   code();
   inter=inter^res1;
   i2=0;
}

 int decode_pc1()
 {
   short c1;
   si=0;
   x1a2=0;
   i2=0;

   crcpuk_temp = 0xffff;

    for (fois=0;fois<=255;fois++)
     {
       c1=bufferc[fois];
       assemble(); 
       cfc=inter>>8;
       cfd=inter&255;  

       c1 = c1 ^ (cfc^cfd);

       for (compte=0;compte<=9;compte++)
	 {
           cle[compte]=cle[compte]^c1;
	 }
       bufferc[fois]=c1;
       crc_pukall(bufferc[fois]);
     }

 }


