/* EmuTO8 -- X display
   Copyright (C) 1998 Jerome Thoen */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef SHM_SUPPORT
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

#include "emuto8.h"
#include "display.h"
#include "memory.h"
#include "console.h"
#include "video.h"

#ifdef SHM_SUPPORT
int use_shm = 0;
#endif

int visible_cols = 2;
int visible_lines = 16;

int size_x;
int size_y;

int bytes_per_pixel;
static int bitdepth;
static int bits_per_pixel;

static Display *display;
static int screen;
static Window rootwin, vidwin;

static int black, white;
static Colormap cmap;

static XVisualInfo visualinfo;
static Visual *visual;
static GC gc;

XImage *img;

unsigned long pixels[16];

XColor xpseudocolors[16];        /* for PseudoColor */

unsigned long xcolors[4096];     /* for TrueColor */

struct charconv {
  unsigned int c;
  char *s;
} keyconv[] = {
  3, "\x03",
  168, "\x16\x48",
  94, "\x16\x43",
  30, "^",
  163, "\x16\x23",
  167, "\x16\x27",
  176, "\x16\x30",
  188, "\x16\x3c",
  189, "\x16\x3d",
  190, "\x16\x3e",
  177, "\x16\x31",
  247, "\x16\x38",
  230, "\x16\x7a",
  198, "\x16\x6a",
  223, "\x16\x7b",
  192, "\x16\x41" "A",
  193, "\x16\x42" "A",
  194, "\x16\x43" "A",
  196, "\x16\x48" "A",
  199, "\x16\x4b" "C",
  200, "\x16\x41" "E",
  201, "\x16\x42" "E",
  202, "\x16\x43" "E",
  203, "\x16\x48" "E",
  204, "\x16\x41" "I",
  205, "\x16\x42" "I",
  206, "\x16\x43" "I",
  207, "\x16\x48" "I",
  210, "\x16\x41" "O",
  211, "\x16\x42" "O",
  212, "\x16\x43" "O",
  214, "\x16\x48" "O",
  217, "\x16\x41" "U",
  218, "\x16\x42" "U",
  219, "\x16\x43" "U",
  220, "\x16\x48" "U",
  224, "\x16\x41" "a",
  225, "\x16\x42" "a",
  226, "\x16\x43" "a",
  228, "\x16\x48" "a",
  231, "\x16\x4b" "c",
  232, "\x16\x41" "e",
  233, "\x16\x42" "e",
  234, "\x16\x43" "e",
  235, "\x16\x48" "e",
  236, "\x16\x41" "i",
  237, "\x16\x42" "i",
  238, "\x16\x43" "i",
  239, "\x16\x48" "i",
  242, "\x16\x41" "o",
  243, "\x16\x42" "o",
  244, "\x16\x43" "o",
  246, "\x16\x48" "o",
  249, "\x16\x41" "u",
  250, "\x16\x42" "u",
  251, "\x16\x43" "u",
  252, "\x16\x48" "u",
  0, "" };
 
struct charconv keysymconv[] = {
  XK_Tab, "\x02",
  XK_BackSpace, "\x7f",
  XK_Return, "\x0d",
  XK_Home, "\x1e",
  XK_Left, "\x08",
  XK_Up, "\x0b",
  XK_Right, "\x09",
  XK_Down, "\x0a",
  XK_Page_Up, "\x0c",
  XK_Delete, "\x1d",
  XK_End, "\x16",
  XK_Insert, "\x1c",
  0, "" };

static XImage *get_image(int w, int h)
{
    XImage *new_img;
    char *p;

#ifdef SHM_SUPPORT
    if (use_shm) {
      XShmSegmentInfo *shminfo = mmalloc(sizeof (XShmSegmentInfo));

      printf("Using MIT-SHM extension.\n");
      new_img = XShmCreateImage(display, visual, bitdepth, ZPixmap, 0, shminfo, w, h);

      shminfo->shmid = shmget(IPC_PRIVATE, h * new_img->bytes_per_line, IPC_CREAT | 0777);
      shminfo->shmaddr = new_img->data = (char *)shmat (shminfo->shmid, 0, 0);
      shminfo->readOnly = False;
      /* let the xserver attach */
      XShmAttach (display, shminfo);
      XSync (display, 0);
      /* now deleting means making it temporary */
      shmctl (shminfo->shmid, IPC_RMID, 0);

      /*
       * finally not needed :)
      if (new_img->bytes_per_line != w * bytes_per_pixel)
	printf("HUUUUUM!\n");
      */

      return new_img;
    }
#endif

    printf("Using normal image buffer.\n");
    p = (char *)mmalloc (h * w * bytes_per_pixel);
    new_img = XCreateImage(display, visual, bitdepth, ZPixmap, 0, p, w, h, 32, 0);

    return new_img;
}

/* bitshift code from UAE */

static __inline__ int bitsInMask(unsigned long mask)
{
    /* count bits in mask */
    int n = 0;
    while(mask) {
        n += mask&1;
        mask >>= 1;
    }
    return n;
}

static __inline__ int maskShift(unsigned long mask)
{
    /* determine how far mask is shifted */
    int n = 0;
    while(!(mask&1)) {
        n++;
        mask >>= 1;
    }
    return n;
}

static int gamma[16] = { 0, 66, 93, 114, 131, 147, 161, 174, 186, 198, 208, 218, 228, 237, 246, 255 };

static unsigned long doMask(int p, int bits, int shift)
{
    unsigned long val = gamma[p] * 0x01010101UL;
    val >>= (32 - bits);
    val <<= shift;

    return val;
}

static void alloc_truecolors(unsigned long red_mask, unsigned long green_mask, unsigned long blue_mask)
{
  int i, red_shift, green_shift, blue_shift, red_bits, green_bits, blue_bits;
  unsigned long r, g, b;

  red_shift = maskShift(red_mask);
  green_shift = maskShift(green_mask);
  blue_shift = maskShift(blue_mask);
  red_bits = bitsInMask(red_mask);
  green_bits = bitsInMask(green_mask);
  blue_bits = bitsInMask(blue_mask);

  for (i = 0; i < 4096; i++) {
    r = doMask(i & 0xf, red_bits, red_shift);
    g = doMask((i >> 4) & 0xf, green_bits, green_shift);
    b = doMask(i >> 8, blue_bits, blue_shift);
    xcolors[i] = r | g | b;
  }
}

void set_pseudocolors(void)
{
  int i;

  for (i = 0; i < 16; i++) {
    xpseudocolors[i].red = gamma[status.colorreg[i*2] & 0xf] * 0x0101;
    xpseudocolors[i].green = gamma[status.colorreg[i*2] >> 4] * 0x0101;
    xpseudocolors[i].blue = gamma[status.colorreg[i*2+1] & 0xf] * 0x0101;
    xpseudocolors[i].flags = DoRed|DoGreen|DoBlue;
  }
  XStoreColors(display, cmap, xpseudocolors, 16);
}

int display_init(char *name)
{
  int i, j;
  XPixmapFormatValues *xpfvs;

  size_x = visible_cols * 16 * 2 + 640;
  size_y = visible_lines * 2 * 2 + 400;

  display = XOpenDisplay(name);
  if(!display) {
    fprintf(stderr, "Can't connect to X server %s\n", XDisplayName(name));
    return 0;
  }

  screen = XDefaultScreen(display);
  rootwin = XRootWindow(display, screen);

  /* bitdepth code based on UAE */

  if (!XMatchVisualInfo(display, screen, 12, TrueColor, &visualinfo))
    if (!XMatchVisualInfo(display, screen, 15, TrueColor, &visualinfo))
      if (!XMatchVisualInfo(display, screen, 16, TrueColor, &visualinfo))
	if (!XMatchVisualInfo(display, screen, 24, TrueColor, &visualinfo))
	  if (!XMatchVisualInfo(display, screen, 8, PseudoColor, &visualinfo)) {
	  fprintf(stderr, "Sorry, visual not supported\n");
	  return 0;
	}

  visual = visualinfo.visual;
  bitdepth = visualinfo.depth;

  xpfvs = XListPixmapFormats(display, &i);
  for (j = 0; j < i && xpfvs->depth != bitdepth; j++, xpfvs++);
 
  bits_per_pixel = xpfvs->bits_per_pixel;
  bytes_per_pixel = bits_per_pixel / 8;

  fprintf(stderr, "Using %d bit visual, %d bits per pixel\n", bitdepth, bits_per_pixel);

  cmap = DefaultColormap(display, screen);
  black = BlackPixel(display, screen);
  white = WhitePixel(display, screen);
  gc = DefaultGC(display, screen);

  if (bytes_per_pixel == 1) {
    if (XAllocColorCells(display, cmap, 0, 0, 0, pixels, 16)) {
      for (i = 0; i < 16; i++) {
	xpseudocolors[i].pixel = pixels[i];
      }
    } else {
      fprintf(stderr, "Can't allocate 16 colors\n");
      return 0;
    }
  } else    
    alloc_truecolors(visualinfo.red_mask, visualinfo.green_mask, visualinfo.blue_mask);

  vidwin = XCreateSimpleWindow(display, rootwin, 0, 0, size_x, size_y, 0, white, black);
  XStoreName(display, vidwin, "EmuTO8");
  XSelectInput(display, vidwin, ExposureMask|KeyPressMask|KeyReleaseMask|ButtonPressMask|ButtonReleaseMask|PointerMotionMask);
  XMapWindow(display, vidwin);

  XFlush(display);

  img = get_image(size_x, size_y);

  return 1;
}

void display_raise_window(void)
{
  XMapRaised(display, vidwin);
  XSetInputFocus(display, vidwin, RevertToNone, CurrentTime);
}

void display_update()
{
#ifdef SHM_SUPPORT
  if (use_shm) {
    XShmPutImage(display, vidwin, gc, img, 0, 0, 0, 0, size_x, size_y, 0);
    XSync (display, 0);
  } else
#endif
    XPutImage(display, vidwin, gc, img, 0, 0, 0, 0, size_x, size_y);
}

__inline__ void display_update_gpl(int x1, int y1)
{
#ifdef SHM_SUPPORT
  if (use_shm)
    XShmPutImage(display, vidwin, gc, img, x1, y1, x1, y1, 16, 16, 0);
  else
#endif
    XPutImage(display, vidwin, gc, img, x1, y1, x1, y1, 16, 16);
}

__inline__ void display_flush(void)
{
  XFlush(display);
}

void decode_key(XKeyEvent *event, char *buffer)
{
  unsigned char inputbuf[2] = " ";
  KeySym k;

  buffer[0] = buffer[1] = buffer[2] = buffer[3] = '\0';

  XLookupString(event, inputbuf, 1, &k, NULL);
  /*printf("%c = %d = %x\n", *inputbuf, *inputbuf, k);*/
  if (*inputbuf >= 0x20 && *inputbuf <= 0x7e && *inputbuf != 94) {
    *buffer = *inputbuf;
    return;
  } else {
    struct charconv *p = keyconv;
    
    while (p->c) {
      if (*inputbuf == p->c) {
	strcpy(buffer, p->s);
	return;
      }
      p++;
    }

    p = keysymconv;

    while (p->c) {
      if (k == p->c) {
	strcpy(buffer, p->s);
	return;
      }
      p++;
    }
  }
}

static int joystick_press(XKeyEvent *event)
{
  KeySym k;
  int r = 1;
  tt_u8 *reg = &status.piabprain;

  k = XLookupKeysym(event, 0);

  switch (k) {
  case XK_KP_Up :
    *reg |= 0x0f;
    *reg &= ~0x01;
    break;
  case XK_KP_Right :
    *reg |= 0x0f;
    *reg &= ~0x08;
    break;
  case XK_KP_Down :
    *reg |= 0x0f;
    *reg &= ~0x02;
    break;
  case XK_KP_Left :
    *reg |= 0x0f;
    *reg &= ~0x04;
    break;
  case XK_KP_Home :
    *reg |= 0x0f;
    *reg &= ~0x05;
    break;
  case XK_KP_Page_Up :
    *reg |= 0x0f;
    *reg &= ~0x09;
    break;
  case XK_KP_Page_Down :
    *reg |= 0x0f;
    *reg &= ~0x0a;
    break;
  case XK_KP_End :
    *reg |= 0x0f;
    *reg &= ~0x06;
    break;
    /*
  case XK_KP_Insert :
    piabprain |= 0x0f;
    piabprbin &= ~0x04;
    break;
    */
  default :
    r = 0;
  }
  return r;
}

static int key_press_event(XKeyEvent *event)
{
  switch (XLookupKeysym(event, 0)) {
  case XK_Control_R :
    status.piabprbin &= ~0x40;
    return 1;
  case XK_Mode_switch :
    status.piabprbin &= ~0x04;
    return 1;
  case XK_Escape :
    activate_console = 1;
    return 1;
  }
  return 0;
}

static void key_release_event(XKeyEvent *event)
{
  switch (XLookupKeysym(event, 0)) {
  case XK_Control_R :
    status.piabprbin |= 0x40;
    break;
  case XK_Mode_switch :
    status.piabprbin |= 0x04;
    break;
  }
}

static void joystick_release(void)
{
  status.piabprain |= 0x0f;
}

void display_handle_events(void)
{
  XEvent event;
  int cont = 1;
  static int joy_timeout = 0, mouse_timeout = 0;
  static char buffer[4] = "   ";

  if (joy_timeout > 0)
    if(--joy_timeout == 0) {
      joystick_release();
    }

  if (mouse_timeout > 0)
    if (--mouse_timeout == 0)
      status.piabprbin |= 0x40;

  while (XPending(display) && cont) {

    XNextEvent(display, &event);

    switch (event.type) {
    case KeyPress :
      /*printf("Keysym:%x\n", XLookupKeysym((XKeyEvent *)&event, 0));*/
      if (key_press_event((XKeyEvent *)&event))
	break;
      if (joystick_press((XKeyEvent *)&event)) {
	cont = 0;
	joy_timeout = 5;
	break;
      }
      decode_key((XKeyEvent *)&event, buffer);
      /*
	printf("%x %x %x\n", buffer[0], buffer[1], buffer[2]);
      */
      if (buffer[0] != '\0') {
	record_key(buffer);
	ktest = 1;
      }
      break;
    case KeyRelease :
      ktest = 0;
      key_release_event((XKeyEvent *)&event);
      break;
    case ButtonPress :
      if (status.sellp)
	status.regE7C3 |= 0x02;
      else
	status.piabprain &= ~(((XButtonEvent *)&event)->button);
      break;
    case ButtonRelease :
      if (status.sellp)
	status.regE7C3 &= ~0x02;
      else
	status.piabprain |= ((XButtonEvent *)&event)->button;
      break;
    case MotionNotify :
      status.lpx = (((XMotionEvent *)&event)->x - visible_cols * 16) / 2;
      status.lpy = (((XMotionEvent *)&event)->y - visible_lines * 2) / 2;
      if (status.lpx < 0)
	status.lpx = 0;
      if (status.lpx > 319)
	status.lpx = 319;
      if (!status.sellp) {
	set_memb(0x60d8, (tt_u8)(status.lpx >> 8));
	set_memb(0x60d9, (tt_u8)status.lpx);
	set_memb(0x60d6, (tt_u8)(status.lpy >> 8));
	set_memb(0x60d7, (tt_u8)status.lpy);
	status.piabprbin &= ~0x40;
	mouse_timeout = 3;
      }
      break;
    case Expose :
      need_border_refresh = need_screen_refresh = 1;
      break;
    }
  }
}









