/* EmuTO8 -- Memory emulation
   Copyright (C) 1998 Jerome Thoen */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#ifdef USE_MAPPED_MEMORY
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "emuto8.h"
#include "emu6809.h"
#include "int6809.h"
#include "disk.h"
#include "memory.h"
#include "console.h"
#include "display.h"
#include "video.h"
#include "zfile.h"

void set_regb(tt_u16 adr, tt_u8 val);
tt_u8 get_regb(tt_u16 adr);

int quickvideo = 1;
int needupdate = 0;
int ktest;
tt_u16 memtrapaddr;
int memtrap = 0;

struct to8_status status = { 0 };

static struct to8_save *save = NULL;

tt_u8 *ramdata;    /* 256 or 512 kb of ram */
static tt_u8 *dummybank;
static tt_u8 *memvidaddr, *memmonaddr, *memramaddr, *memromaddr;

#ifdef USE_MAPPED_MEMORY
static tt_u8 *mem_image;
static tt_u8 *mem_rom;
static tt_u8 *mem_vid;
static tt_u8 *mem_ram;
static tt_u8 *mem_mon;
static int mem_fd;
#endif

static const char *files[] = { "b512_b0.rom",
			       "b512_b1.rom",
			       "basic1.rom",
			       "fichier.rom",
			       "to8mon1.rom",
			       "to8mon2.rom" };

static const size_t romsizes[] = { 16384, 16384, 16384, 16384, 8192, 8192 };

static tt_u8 *extroms[2] = { NULL };
static tt_u8 *roms[6];

#ifdef USE_MAPPED_MEMORY
void calc_memaddr(void)
{
  tt_u8 *newaddr;

  if (status.selrama)
    newaddr = ramdata + 0x2000;
  else
    newaddr = ramdata;
  if (mem_vid != newaddr) {
    mmap(mem_image + 0x4000, 0x2000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)newaddr);
    mem_vid = newaddr;
  }

  if (status.bank <= 15 || status.ramext)
    newaddr = ramdata + status.bank * 0x4000;
  else
    newaddr = dummybank;
  if (mem_ram != newaddr) {
    mmap(mem_image + 0xa000, 0x4000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)newaddr);
    mem_ram = newaddr;
  }

  newaddr = roms[status.monbank];
  if (mem_mon != newaddr) {
    mmap(mem_image + 0xe000, 0x2000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)newaddr);
    mem_mon = newaddr;
  }

  if (btst(status.regE7E6, 0x20)) {
    if ((status.regE7E6 & 0x1f) <= 15 || status.ramext)
      newaddr = ramdata + (status.regE7E6 & 0x1f) * 0x4000;
    else
      newaddr = dummybank;
    if (mem_rom != newaddr) {
      mmap(mem_image, 0x2000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)(newaddr + 0x2000));
      mmap(mem_image + 0x2000, 0x2000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)newaddr);
      mem_rom = newaddr;
    }
  } else {
    if (status.selcart)
      if (status.card)
	newaddr = extroms[status.rombank & 0x1];
      else
	newaddr = dummybank;
    else
      newaddr = roms[status.rombank];
    if (mem_rom != newaddr) {
      mmap(mem_image, 0x4000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)newaddr);
      mem_rom = newaddr;
    }
  }
  memvidaddr = mem_vid - 0x4000;
  memmonaddr = mem_mon - 0xe000;
  memramaddr = mem_ram - 0xa000;
  memromaddr = mem_rom;
}
#else
void calc_memaddr(void)
{
  if (status.selrama)
    memvidaddr = ramdata + 0x2000 - 0x4000;
  else
    memvidaddr = ramdata - 0x4000;
  if (status.bank <= 15 || status.ramext)
    memramaddr = ramdata + status.bank * 0x4000 - 0xa000;
  else
    memramaddr = dummybank - 0xa000;
  memmonaddr = roms[status.monbank] - 0xe000;
  if (btst(status.regE7E6, 0x20))
    if ((status.regE7E6 & 0x1f) <= 15 || status.ramext)
      memromaddr = ramdata + (status.regE7E6 & 0x1f) * 0x4000;
    else
      memromaddr = dummybank;
  else
    if (status.selcart)
      if (status.card)
	memromaddr = extroms[status.rombank & 0x1];
      else
	memromaddr = dummybank;
    else
      memromaddr = roms[status.rombank];
}
#endif

int load_rom(char *romfile)
{
  int s, fd;
  FILE *f;

  f = fopen(romfile, "rb");
  if (!f) {
    printf("ROM card %s not found\n", romfile);
    return 0;
  }
  if (!extroms[0]) {
    printf("Allocating memory for ROM card\n");
#ifdef USE_MAPPED_MEMORY
    fd = open("/dev/zero", O_RDWR);
    extroms[0] = mmap(0, 0x4000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
    extroms[1] = mmap(0, 0x4000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
    close(fd);
    if (extroms[0] == -1 || extroms[1] == -1) {
      fclose(f);
      return 0;
    }
#else      
    extroms[0] = (tt_u8 *)malloc(0x4000);
    extroms[1] = (tt_u8 *)malloc(0x4000);
    if (!extroms[0] || !extroms[1]) {
      fclose(f);
      return 0;
    }
#endif
  }
  printf("Loading ROM cartridge %s\n", romfile);
  s = fread(extroms[0], 1, 0x4000, f);
  if (s == 0x1000) {
    printf("4 kb ROM loaded\n");
    memcpy(extroms[0] + 0x1000, extroms[0], 0x1000);
    memcpy(extroms[0] + 0x2000, extroms[0], 0x2000);
    memcpy(extroms[1], extroms[0], 0x4000);
    status.card = 1;
  } else if (s == 0x2000) {
    printf("8 kb ROM loaded\n");
    memcpy(extroms[0] + 0x2000, extroms[0], 0x2000);
    memcpy(extroms[1], extroms[0], 0x4000);
    status.card = 1;
  } else if (s == 0x4000) {
    s = fread(extroms[1], 1, 0x4000, f);
    if (s == 0) {
      printf("16 kb ROM loaded\n");
      memcpy(extroms[1], extroms[0], 0x4000);
      status.card = 1;
    } else if (s == 0x4000) {
      printf("32 kb ROM loaded\n");
      status.card = 1;
    } else
      status.card = 0;
  } else
    status.card = 0;

  fclose(f);

  if (status.card == 0)
    printf("Error while loading ROM file\n");

  calc_memaddr();

  return status.card;
}

int memory_init(char *romdirpath, char *romfile)
{
  int i, fd;
  FILE *f;
  char filename[1024];
  size_t ramsize = status.ramext ? 0x80000 : 0x40000;

  status.monbank = 4;
  status.piaaprb_datamode = status.piabprb_datamode = status.piabpra_datamode = 1;
  status.piabprain = status.piabprbin = 0xff;
  status.lpy = -1;

#ifdef USE_MAPPED_MEMORY
  fd = open("/dev/zero", O_RDWR);
  mem_fd = open("/proc/self/mem", O_RDWR);
#endif

#ifdef USE_MAPPED_MEMORY
  dummybank = mmap(0, 0x4000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
  if (dummybank < 0)
    abort();
#else
  dummybank = (tt_u8 *)mmalloc(0x4000);
#endif
  memset(dummybank, 0xff, 0x4000);

#ifdef USE_MAPPED_MEMORY
  mem_image = mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
  if (mem_image < 0) {
    fprintf(stderr, "Not enough virtual memory\n");
    return 0;
  }
#endif

  printf("Allocating %d bytes for RAM\n", ramsize);
#ifdef USE_MAPPED_MEMORY
  ramdata = mmap(0, ramsize, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
  if (ramdata < 0)
    abort();
#else
  ramdata = (tt_u8 *)mmalloc(ramsize);
#endif

  for(i = 0; i < 6; i++) {
    printf("Allocating %d bytes for ROM %s\n", romsizes[i], files[i]);
#ifdef USE_MAPPED_MEMORY
    roms[i] = mmap(0, romsizes[i], PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
    if (roms[i] < 0)
      abort();
#else
    roms[i] = (tt_u8 *)mmalloc(romsizes[i]);
#endif
    strcpy(filename, romdirpath);
    strcat(filename, files[i]);
    f = fopen(filename, "rb");
    if (!f) {
      f = fopen(files[i], "rb");
      if (!f)
	fprintf(stderr, "ROM file %s not found.\n", files[i]);
      return 0;
    }
    if(fread(roms[i], 1, romsizes[i], f) != romsizes[i]) {
      fprintf(stderr, "Error while loading ROM file %s.\n", files[i]);
      fclose(f);
      return 0;
    }
    fclose(f);
  }

  if (romfile[0] != '\0')
    load_rom(romfile);
  else {
    f = fopen("user.rom", "rb");
    if (f) {
      fclose(f);
      load_rom("user.rom");
    }
  }

#ifdef USE_MAPPED_MEMORY
  mmap(mem_image + 0x6000, 0x4000, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_PRIVATE, mem_fd, (off_t)(ramdata + 0x4000));
  close(fd);
  mem_rom = mem_vid = mem_ram = mem_mon = (tt_u8 *)(-1);
#endif

  calc_memaddr();

  return 1;
}

#ifdef USE_MAPPED_MEMORY
__inline__ tt_u8 get_memb(tt_u16 adr)
{
#ifdef MEM_TRAP
  if (memtrap)
    if (adr == memtrapaddr) {
      activate_console = 1;
      memtrap = 0;
    }
#endif
  if (adr <= 0xe7af || adr >= 0xe800) 
    return mem_image[adr];
  else
    return get_regb(adr);
}
#else
tt_u8 get_memb(tt_u16 adr)
{
#ifdef MEM_TRAP
  if (memtrap)
    if (adr == memtrapaddr) {
      activate_console = 1;
      memtrap = 0;
    }
#endif
  if (adr <= 0x3fff) { 
    /* user rom */
    if (btst(status.regE7E6, 0x20))
      return memromaddr[adr ^ 0x3000];
    else
      return memromaddr[adr];
  }
  if (adr <= 0xdfff) {
    if (adr <= 0x5fff) {   
      return memvidaddr[adr];
    }
    if (adr <= 0x9fff) {   
      return ramdata[(int)adr - 0x6000 + 0x4000];
    }
    if (adr <= 0xdfff) {   
      /* user ram */
      return memramaddr[adr];
    }
  }
  if (adr <= 0xe7af || adr >= 0xe800) {
    /* monitor */
    return memmonaddr[adr];
  }
  /* registers */
  return get_regb(adr);
}
#endif

__inline__ tt_u16 get_memw(tt_u16 adr)
{
  return (tt_u16)get_memb(adr) << 8 | (tt_u16)get_memb(adr + 1);
}

#ifdef USE_MAPPED_MEMORY
void set_memb(tt_u16 adr, tt_u8 val)
{
#ifdef MEM_TRAP
  if (memtrap)
    if (adr == memtrapaddr) {
      activate_console = 1;
      memtrap = 0;
    }
#endif

  switch (adr >> 12) {
  case 0x0 : case 0x1 : case 0x2 : case 0x3 :
    if (btst(status.regE7E6, 0x20)) { /* rom masked */
      if (btst(status.regE7E6, 0x40)) /* write allowed */
	if (status.romrampage == status.vidpage && quickvideo) {
	  tt_u16 nadr = adr & ~0x2000;
	  tt_u8 oldval = mem_image[adr];

	  mem_image[adr] = val;
	  if (nadr <= 0x1f3f && oldval != val)
	    quick_decode_gpl(nadr);
	} else {
	  if (status.romrampage <= 15 || status.ramext)
	    mem_image[adr] = val;
	}
    } else {
      if (adr <= 0x7) {
	status.rombank = adr & 0x3;
	calc_memaddr();
      }
    }
    return;
  case 0x4 : case 0x5 :
    if (adr <= 0x5f3f && status.vidpage == 0 && quickvideo) {
      tt_u8 oldval = mem_image[adr];

      mem_image[adr] = val;
      if (oldval != val)
	quick_decode_gpl(adr - 0x4000);
    } else {
      mem_image[adr] = val;
    }
    return;
  case 0x6 : case 0x7 : case 0x8 : case 0x9 :
    if (status.vidpage == 1 && quickvideo) {
      tt_u16 nadr = (adr & ~0x2000) - 0x6000;

      if (nadr <= 0x1f3f) {
	tt_u8 oldval = mem_image[adr];

	if (oldval != val)
	  quick_decode_gpl(nadr);
      }
    }
    mem_image[adr] = val;
    return;
  case 0xa : case 0xb : case 0xc : case 0xd :
    if (status.bank == status.vidpage && quickvideo) {
      tt_u16 nadr = (adr & ~0x2000) - 0xa000;

      if (nadr <= 0x1f3f) {
	tt_u8 oldval = mem_image[adr];

	if (oldval != val)
	  quick_decode_gpl(nadr);
      }
    }
    if (status.bank <= 15 || status.ramext)
      mem_image[adr] = val;
    return;
  default :
    if (adr >= 0xe7b0 && adr <= 0xe7ff)
      set_regb(adr, val);
    return;
  }
}
#else
void set_memb(tt_u16 adr, tt_u8 val)
{
  int nadr;
  tt_u8 oldval;

#ifdef MEM_TRAP
  if (memtrap)
    if (adr == memtrapaddr) {
      activate_console = 1;
      memtrap = 0;
    }
#endif
  if (adr <= 0x3fff) {   /* 0000-3FFF */
    /* user rom */
    set_romb(adr, val);
    return;
  }
  if (adr <= 0x5fff) {   /* 4000-5FFF */
    oldval = memvidaddr[adr];
    memvidaddr[adr] = val;
    if (quickvideo && adr <= 0x5f3f && oldval != val)
      quick_decode_gpl(adr - 0x4000);
    return;
  }
  if (adr <= 0x9fff) {   /* 6000-9FFF */
    ramdata[(int)adr - 0x6000 + 0x4000] = val;
    return;
  }
  if (adr <= 0xdfff) {   /* A000-DFFF */
    /* user ram */
    if (status.bank <= 15 || status.ramext)
      memramaddr[adr] = val;
    return;
  }
  if (adr <= 0xe7af) {   /* E000-E7AF */
    /* monitor */
    /* nowrite(adr, val); */
    return;
  }
  if (adr <= 0xe7ff) {   /* E7B0-E7FF */
    /* registers */
    set_regb(adr, val);
    return;
  }   /* E800-FFFF */
    /* monitor */
    /* nowrite(adr, val); */
    
}
#endif

void set_memw(tt_u16 adr, tt_u16 val)
{
  set_memb(adr, (tt_u8)(val >> 8));
  set_memb(adr + 1, (tt_u8)val);
}

void set_romb(tt_u16 adr, tt_u8 val)
{
  if (btst(status.regE7E6, 0x20) && btst(status.regE7E6, 0x40)) {
    /* rom masked and write allowed */
    memromaddr[adr ^ 0x3000] = val;
  } else {
    if (adr <= 0x7) {
      status.rombank = adr & 0x3;
      calc_memaddr();
    }
  }
}
 
tt_u8 get_monb(tt_u16 adr)
{
  return roms[status.monbank][adr - 0xe000];
}

/* hardware registers */

tt_u8 get_regb(tt_u16 adr)
{
  tt_u8 val;

  switch(adr & 0xff) {

    /* 6846 */

  case 0xc0 :
  case 0xc4 :
    /* CSR */
    if (status.timer_irq)
      status.timer_readirq = 1;
    return status.timer_irq ? 0x81 : 0x00;
    break;
  case 0xc1 :
    /* PCR */
    return status.regE7C1;
  case 0xc2 :
    /* DDR */
    return 0x3d;
  case 0xc3 :
    /* PDR */
    return status.regE7C3;
  case 0xc5 :
    /* TCR */
    return status.regE7C5;
  case 0xc6 :
    /* TMSB*/
    if (status.timer_readirq) {
      status.timer_irq = 0;
      status.timer_readirq = 0;
    }
    return (tt_u8)(status.timer_val >> 8);
  case 0xc7 :
    /* TLSB*/
    return (tt_u8)status.timer_val;

    /* 6821 A */

  case 0xc8 :
    /* piaa pra */
    return ktest & 0x01;
    break;
  case 0xc9 :
    /* piaa prb */
    break;
  case 0xca :
    /* piaa cra */
    break;
  case 0xcb :
    /* piaa crb */
    return status.piaaprb_datamode ? 0x04 : 0x00;
    break;

    /* 6821 B */

  case 0xcc :
    /* data piab pra */
    if (status.piabpra_datamode)
      return (status.piabprain & ~status.piabpramask) | (status.piabpraout & status.piabpramask);
    else
      return status.piabpramask;
    break;
  case 0xcd :
    /* data piab prb */
    if (status.piabprb_datamode)
      return (status.piabprbin & ~status.piabprbmask) | (status.piabprbout & status.piabprbmask);
    else
      return status.piabprbmask;
    break;
  case 0xce :
    /* control piab cra */
    return status.piabcra;
    break;
  case 0xcf :
    /* control piab crb */
    return status.piabcrb;
    break;

  case 0xd0 :
    return 0x80;
  case 0xd1 :
    return 0x30;
  case 0xda :
    /* palette data */
    val = status.colorreg[status.regE7DB];
    status.regE7DB++;
    if (status.regE7DB > 31)
    status.regE7DB = 0;
    return val;
  case 0xdb :
    return status.regE7DB;
    break;
  case 0xe4 :
    if (status.selreglp)
      return 0;
    else
      return status.regE7DD & 0xc0;
    break;
  case 0xe5 :
    if (status.selreglp)
      return 0;
    else
      return status.bank & 0x1f;
    break;
  case 0xe6 :
    if (status.selreglp)
      return 0;
    else
      return status.regE7E6;
    break;
  case 0xe7 :
    return giveE7E7() | status.selreglp;
    break;
  default :
    if (debugio) {
      printf("read access at %04hX\n", adr);
      activate_console = 1;
    }
  }

  return 0;
}

void set_regb(tt_u16 adr, tt_u8 val)
{
  switch(adr & 0xff) {

    /* 6846 */

  case 0xc1 :
    /* PCR */
    status.regE7C1 = val;
    if ((val & 0x30) == 0x30)
      status.mute = btst(val, 0x08);
    break;
  case 0xc2 :
    /* DDR */
    break;
  case 0xc3 :
    /* PDR */
    {
      int oldselrama = status.selrama;
      int oldselcart = status.selcart;
      int oldmonbank = status.monbank;

      status.regE7C3 = val;
      status.selrama = btst(val, 0x01);
      status.selcart = !btst(val, 0x04);
      status.monbank = btst(val, 0x10) ? 5 : 4;

      if (oldselrama != status.selrama || oldselcart != status.selcart || oldmonbank != status.monbank)
	calc_memaddr();
      break;
    }
  case 0xc5 :
    /* TCR */
    status.regE7C5 = val;
    /*    printf("E7C5 TCR = %s\n", bin8str(val)); */
    status.timer_mode = val >> 3 & 0x3;
    status.timer_enabled = !btst(val, 0x01);
    if (!status.timer_enabled) {
      /* printf("timer reset at %04X\n", timer_val_init); */
      status.timer_val = status.timer_val_init;
      status.timer_irq = 0;
    }
    break;
  case 0xc6 :
    /* TMSB*/
    status.timer_val_init = (tt_u16)val << 8;
    break;
  case 0xc7 :
    /* TLSB*/
    status.timer_val_init |= val;
    /*    printf("TLSB write timer latch = %04X\n", timer_val_init);*/
    if (status.timer_mode == 0) {
      /*      printf("TLSB restarting\n"); */
      status.timer_val = status.timer_val_init;
      status.timer_irq = 0;
      status.timer_enabled = 1;
    }
    break;

  case 0xc8 :
    /* piaa pra */
    break;
  case 0xc9 :
    /* piaa prb */
    if (!status.piaaprb_datamode && status.rampia) {
      int oldbank = status.bank;

      switch(val & 0xf8) {
      case 0x08 :
	status.bank = 2;
	break;
      case 0x10 :
	status.bank = 3;
	break;
      case 0xe0 :
	status.bank = 4;
	break;
      case 0xa0 : 
	status.bank = 5;
	break;
      case 0x60 : 
	status.bank = 6;
	break;
      case 0x20 :
	status.bank = 7;
	break;
      default :
	printf("E7C9 = %02X !\n", val);
	if (debugio)
	  activate_console = 1;
      }
      if (oldbank != status.bank)
	calc_memaddr();
    }
    break;
  case 0xca :
    /* control piaa cra */
    break;
  case 0xcb :
    /* control piaa crb */
    status.piaaprb_datamode = btst(val, 0x04);
    break;
  case 0xcc :
    /* data piab pra */
    if (status.piabpra_datamode)
      status.piabpraout = val;
    else
      status.piabpramask = val;
    break;
  case 0xcd :
    /* data piab prb */
    if (status.piabprb_datamode)
      status.piabprbout = val;
    else
      status.piabprbmask = val;
    break;
  case 0xce :
    /* control piab cra */
    status.piabpra_datamode = btst(val, 0x04);
    status.piabcra = val;
    break;
  case 0xcf :
    /* control piab crb */
    status.piabprb_datamode = btst(val, 0x04);
    status.piabcrb = val;
    break;
  case 0xda :
    /* palette data set */
    status.colorreg[status.regE7DB] = val;
    if (bytes_per_pixel == 1) {
      need_palette_refresh = 1;
    } else {
      need_screen_refresh = 1;
      need_border_refresh = 1;
      if (status.regE7DB % 2) { /* blue */
	status.truecolors[status.regE7DB / 2] = xcolors[status.colorreg[status.regE7DB - 1] | ((tt_u16)(val & 0xf) << 8)];
      } else { /* green red */
	status.truecolors[status.regE7DB / 2] = xcolors[val | ((tt_u16)(status.colorreg[status.regE7DB + 1] & 0xf) << 8)];
      }
    }
    status.regE7DB++;
    if (status.regE7DB > 31)
      status.regE7DB = 0;
    break;
  case 0xdb :
    /* palette adr */
    status.regE7DB = val & 0x1f;
    break;
  case 0xdc : 
    {
      tt_u8 oldreg = status.regE7DC;
      
      status.regE7DC = val;
      if (oldreg != val)
	need_screen_refresh = 1;
      break;
    }
  case 0xdd :
    {
      int oldvidpage = status.vidpage;
      tt_u8 oldreg = status.regE7DD & 0xf;

      status.regE7DD = val;
      status.vidpage = val >> 6 & 0x3;
      if (oldreg != val & 0xf)
	need_border_refresh = 1;
      if (oldvidpage != status.vidpage)
	need_screen_refresh = 1;
      break;
    }
  case 0xe4 :
    status.selreglp = val & 0x01;
    break;
  case 0xe5 :
    if (!status.rampia) {
      int oldbank = status.bank;

      status.bank = val & 0x1f;
      if (oldbank != status.bank)
	calc_memaddr();
    }
    break;
  case 0xe6 :
    {
      tt_u8 oldreg = status.regE7E6;

      status.regE7E6 = val;
      status.romrampage = val & 0x1f;
      if (status.regE7E6 != oldreg)
        calc_memaddr();
      break;
    }
  case 0xe7: 
    if (val & 0x10)
      status.rampia = 0;
    else
      status.rampia = 1;
    break;
  default :
    if (debugio) {
      printf("write access %04hX = %02X (%s) ignored\n", adr, val, bin8str(val));
      activate_console = 1;
    }
  }
}

void mem_dumpstate(void)
{
  int i;

  printf("Ram bank: %d  Rom bank: %d  Mon bank: %d  Video ram: %c  Ram PIA: %c\n", status.bank, status.rombank, btst(status.regE7C3, 0x10), status.selrama ? 'A' : 'B', status.rampia ? 'Y' : 'N');

  printf("E7C0: %s  ", bin8str(get_regb(0xe7c0)));
  printf("E7C1: %s  ", bin8str(status.regE7C1));
  printf("E7C3: %s  ", bin8str(status.regE7C3));
  printf("E7C5: %s  ", bin8str(status.regE7C5));
  printf("E7C6-E7C7: %04hX\n", status.timer_val);
  printf("E7DC: %s  ", bin8str(status.regE7DC));
  printf("E7DD: %s\n", bin8str(status.regE7DD));
  printf("E7E4: %s  ", bin8str(get_regb(0xe7e4)));
  printf("E7E5: %s  ", bin8str(get_regb(0xe7e5)));
  printf("E7E6: %s  ", bin8str(status.regE7E6));
  printf("E7E7: %s\n", bin8str(get_regb(0xe7e7)));

  printf("Color regs:\n");
  for (i = 0; i < 32; i += 2) {
    printf("%02X%02X ", status.colorreg[i + 1], status.colorreg[i]);
    if (i == 14 || i == 30)
      putchar('\n');
  }
  video_dumpstate();
}

void record_key(char *buffer)
{
  tt_u8 c;
  tt_u8 nptr = get_memb(0x6067);
  tt_u8 size = get_memb(0x607b);
  tt_u16 p = get_memw(0x6079);

  while (c = *buffer++) {
    set_memb(p + nptr, c);
    nptr++;
    if (nptr >= size)
      nptr = 0;
  }
  set_memb(0x6067, nptr);
}

void timer_timeout(void)
{
  status.timer_val = status.timer_val_init;
  if (status.regE7C5 & 0x40) {
    status.timer_irq = 1;
    irq();
  }
}

void install_trap(void)
{
  roms[4][0x00fe] = 0x02;
  roms[4][0x03a7] = 0x02;
  roms[4][0x0177] = 0x02;
  roms[5][0x0fc1] = 0x02;
  if (roms[4][0x000f] == 0xa8) /* TO8D */
    roms[4][0x1a60] = 0x02;
  else /* TO8 */
    roms[4][0x1a64] = 0x02;
  roms[4][0x0c30] = 0x02;
}

void trap(void)
{
  switch (rpc-1) {
  case 0xe0fe :
    InitDiskCtrl();
    rts();
    return;
  case 0xe3a7 :
    ReadSector();
    putchar('.');
    fflush(stdout);
    rts();
    return;
  case 0xe177 :
    WriteSector();
    rts();
    return;
  case 0xefc1 :
    if (get_memb(0x6019) & 0x02) {
      set_memb(0x6019, get_memb(0x6019) & ~0x02);
      rts();
    } else
      pshs();
    return;
  case 0xfa60 :
  case 0xfa64 :
    status.sellp = 1;
    if (status.lpy < 0 || status.lpy >= 200)
      setcc(getcc() | 0x01);
    else {
      setcc(getcc() & ~0x01);
      rx = status.lpx;
      ry = status.lpy;
    }
    rts();
    return;
  case 0xec30 :
    status.sellp = 0;
    pshs();
    return;
  default :
    null();
  }
}
    
void memory_reset(void)
{
  set_memb(0xe7dc, 0);
  set_memb(0xe7dd, 0);
  set_memb(0xe7e7, 0);
  set_memb(0xe7e6, 0);
  set_memb(0xe7e5, 0);
  set_memb(0xe7c5, 0);
  status.piabprain = status.piabprbin = status.piabpraout = status.piabprbout = 0xff;
  status.piabpramask = status.piabprbmask = status.piabcra = status.piabcrb = 0x0;
  status.selreglp = 0;
  if (status.ramext)
    memset(ramdata, 0, 0x80000);
  else
    memset(ramdata, 0, 0x40000);
}

int memory_save(char *filename)
{
  FILE *fp;
  char name[1024];
  
  if (!save)
    save = (struct to8_save *)mmalloc(sizeof(struct to8_save));

  if(!strchr(filename, '/')) {
    strcpy(name, "sav/");
    strcat(name, filename);
  } else
    strcpy(name, filename);

  if(!strstr(name, ".sav"))
    strcat(name, ".sav");
	
  if (fp = fopen(name, "rb")) {
    char name2[1024];
    FILE *fp2;

    strcpy(name2, name);
    strcat(name2, ".bak");
    if (fp2 = fopen(name2, "wb")) {
      int c;

      /* hum buffered version needed :) */
      while ((c = getc(fp)) != EOF) 
	putc(c, fp2);

      fclose(fp2);
    }
    fclose(fp);
  }

  fp = fopen(name, "wb");
  if (!fp)
    return 0;

  save->rpc = rpc;
  save->rx = rx;
  save->ry = ry;
  save->ru = ru;
  save->rs = rs;
  save->ra = ra;
  save->rb = rb;
  save->rdp = rdp;
  save->rcc = getcc();
  memcpy(&save->status, &status, sizeof(struct to8_status));

  fwrite(save, sizeof(struct to8_save), 1, fp);
  fwrite(ramdata, status.ramext ? 0x80000 : 0x40000, 1, fp);
  if (status.card) {
    fwrite(extroms[0], 16384, 1, fp);
    fwrite(extroms[1], 16384, 1, fp);
  }

  fclose(fp);

  return 1;
}

int memory_load(char *filename)
{
  FILE *fp;
  int fd;
  char name[1024];

  if (!save)
    save = (struct to8_save *)mmalloc(sizeof(struct to8_save));

  if(!strchr(filename, '/')) {
    strcpy(name, "sav/");
    strcat(name, filename);
  } else
    strcpy(name, filename);

  if(!strstr(name, ".sav"))
    strcat(name, ".sav");

  fp = fopen(name, "rb");
  if (!fp)
    return 0;

  fread(save, sizeof(struct to8_save), 1, fp);

  rpc = save->rpc;
  rx = save->rx;
  ry = save->ry;
  ru = save->ru;
  rs = save->rs;
  ra = save->ra;
  rb = save->rb;
  rdp = save->rdp;
  setcc(save->rcc);

  if (save->status.ramext && !status.ramext) {
    printf("Sorry, you must start emuto8 with option \"-e\" to load this file\n");
    fclose(fp);
    return 0;
  }
  memcpy(&status, &save->status, sizeof(struct to8_status));

  fread(ramdata, status.ramext ? 0x80000 : 0x40000, 1, fp);
  if (status.card) {
    if (!extroms[0]) {
      printf("Allocating memory for ROM card\n");
#ifdef USE_MAPPED_MEMORY
      fd = open("/dev/zero", O_RDWR);
      extroms[0] = mmap(0, 0x4000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
      extroms[1] = mmap(0, 0x4000, PROT_READ|PROT_WRITE, MAP_PRIVATE, fd, 0);
      close(fd);
      if (extroms[0] == -1 || extroms[1] == -1) {
	fclose(fp);
	return 0;
      }
#else      
      extroms[0] = (tt_u8 *)mmalloc(0x4000);
      extroms[1] = (tt_u8 *)mmalloc(0x4000);
#endif
    }
    fread(extroms[0], 16384, 1, fp);
    fread(extroms[1], 16384, 1, fp);
  }
  
  fclose(fp);

  calc_memaddr();

  if (bytes_per_pixel == 1)
    need_palette_refresh = 1;
  else
    need_screen_refresh = need_border_refresh = 1;

  return 1;
}
  


