/* EmuTO8 -- video emulation
   Copyright (C) 1998 Jerome Thoen */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <X11/Xlib.h>

#include "emuto8.h"
#include "display.h"
#include "memory.h"
#include "console.h"
#include "audio.h"
#include "video.h"

enum { 
  INVISIBLE_UP,
  INVISIBLE_UP_LEFT,
  VISIBLE_UP,
  INVISIBLE_UP_RIGHT,
  INVISIBLE_LEFT,
  VISIBLE_LEFT,
  VISIBLE_CENTER,
  VISIBLE_RIGHT,
  INVISIBLE_RIGHT,
  INVISIBLE_DOWN_LEFT,
  VISIBLE_DOWN,
  INVISIBLE_DOWN_RIGHT,
  INVISIBLE_DOWN };

static int cyclecnt = 0;
static int timercnt = 8;
static int soundcnt = 91;

static long frames = 0;

static char updategpl[1000] = {0};

int need_border_refresh = 0;
int need_screen_refresh = 0;
int need_palette_refresh = 0;

static void (*gpl_func)(tt_u8 *d, char *b);
static void (*gpl_border_func)(char *b);
static void (*video_mode_func)(void);

static void gpl_border_bpp1(char *bptr)
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  unsigned char col = pixels[status.regE7DD & 0xf]; 

  for(i = 0; i < 8; i++) {
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    PUTPIXEL_BPP1(bptr2, col);
  }
}

static void gpl_border_bpp2(char *bptr)
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  unsigned short col = status.truecolors[status.regE7DD & 0xf];

  for(i = 0; i < 8; i++) {
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    PUTPIXEL_BPP2(bptr2, col);
  }
}

static void gpl_border_bpp3(char *bptr)
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  unsigned long col = status.truecolors[status.regE7DD & 0xf];

  for(i = 0; i < 8; i++) {
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    PUTPIXEL_BPP3(bptr2, col);
  }
}

static void gpl_border_bpp4(char *bptr)
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  unsigned long col = status.truecolors[status.regE7DD & 0xf];

  for(i = 0; i < 8; i++) {
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    PUTPIXEL_BPP4(bptr2, col);
  }
}

static void gpl_mode0_bpp1(tt_u8 *dptr, char *bptr) /* TO7/70 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned char fc = pixels[ramb >> 3 & 0x0f ^ 0x08];
  unsigned char bc = pixels[(ramb & 0x07) | (ramb & 0x80 ? 0 : 0x08)];
  unsigned char col;

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? fc : bc;
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    PUTPIXEL_BPP1(bptr2, col);
    rama <<= 1;
  }
}

static void gpl_mode0_bpp2(tt_u8 *dptr, char *bptr) /* TO7/70 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned short fc = status.truecolors[ramb >> 3 & 0x0f ^ 0x08];
  unsigned short bc = status.truecolors[(ramb & 0x07) | (ramb & 0x80 ? 0 : 0x08)];
  unsigned short col;

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? fc : bc;
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    PUTPIXEL_BPP2(bptr2, col);
    rama <<= 1;
  }
}

static void gpl_mode0_bpp3(tt_u8 *dptr, char *bptr) /* TO7/70 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long fc = status.truecolors[ramb >> 3 & 0x0f ^ 0x08];
  unsigned long bc = status.truecolors[(ramb & 0x07) | (ramb & 0x80 ? 0 : 0x08)];
  unsigned long col;

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? fc : bc;
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    PUTPIXEL_BPP3(bptr2, col);
    rama <<= 1;
  }
}

static void gpl_mode0_bpp4(tt_u8 *dptr, char *bptr) /* TO7/70 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long fc = status.truecolors[ramb >> 3 & 0x0f ^ 0x08];
  unsigned long bc = status.truecolors[(ramb & 0x07) | (ramb & 0x80 ? 0 : 0x08)];
  unsigned long col;

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? fc : bc;
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    PUTPIXEL_BPP4(bptr2, col);
    rama <<= 1;
  }
}

static void gpl_mode1_bpp1(tt_u8 *dptr, char *bptr) /* bitmap 4 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned char col;

  for(i = 0; i < 8; i++) {
    col = pixels[((ramb & 0x80) >> 7) + ((rama & 0x80) >> 6)];
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    PUTPIXEL_BPP1(bptr2, col);
    rama <<= 1;
    ramb <<= 1;
  }
}

static void gpl_mode1_bpp2(tt_u8 *dptr, char *bptr) /* bitmap 4 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned short col;

  for(i = 0; i < 8; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((rama & 0x80) >> 6)];
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    PUTPIXEL_BPP2(bptr2, col);
    rama <<= 1;
    ramb <<= 1;
  }
}

static void gpl_mode1_bpp3(tt_u8 *dptr, char *bptr) /* bitmap 4 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  for(i = 0; i < 8; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((rama & 0x80) >> 6)];
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    PUTPIXEL_BPP3(bptr2, col);
    rama <<= 1;
    ramb <<= 1;
  }
}

static void gpl_mode1_bpp4(tt_u8 *dptr, char *bptr) /* bitmap 4 */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  for(i = 0; i < 8; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((rama & 0x80) >> 6)];
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    PUTPIXEL_BPP4(bptr2, col);
    rama <<= 1;
    ramb <<= 1;
  }
}

static void gpl_mode2_bpp1(tt_u8 *dptr, char *bptr) /* bitmap 16 */
{
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned char col;

  col = pixels[rama >> 4];
  PUTPIXEL_BPP1_4(bptr, col);
  PUTPIXEL_BPP1_4(bptr2, col);
  col = pixels[rama & 0x0f];
  PUTPIXEL_BPP1_4(bptr, col);
  PUTPIXEL_BPP1_4(bptr2, col);
  col = pixels[ramb >> 4];
  PUTPIXEL_BPP1_4(bptr, col);
  PUTPIXEL_BPP1_4(bptr2, col);
  col = pixels[ramb & 0x0f];
  PUTPIXEL_BPP1_4(bptr, col);
  PUTPIXEL_BPP1_4(bptr2, col);
}

static void gpl_mode2_bpp2(tt_u8 *dptr, char *bptr) /* bitmap 16 */
{
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned short col;

  col = status.truecolors[rama >> 4];
  PUTPIXEL_BPP2_4(bptr, col);
  PUTPIXEL_BPP2_4(bptr2, col);
  col = status.truecolors[rama & 0x0f];
  PUTPIXEL_BPP2_4(bptr, col);
  PUTPIXEL_BPP2_4(bptr2, col);
  col = status.truecolors[ramb >> 4];
  PUTPIXEL_BPP2_4(bptr, col);
  PUTPIXEL_BPP2_4(bptr2, col);
  col = status.truecolors[ramb & 0x0f];
  PUTPIXEL_BPP2_4(bptr, col);
  PUTPIXEL_BPP2_4(bptr2, col);
}

static void gpl_mode2_bpp3(tt_u8 *dptr, char *bptr) /* bitmap 16 */
{
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  col = status.truecolors[rama >> 4];
  PUTPIXEL_BPP3_4(bptr, col);
  PUTPIXEL_BPP3_4(bptr2, col);
  col = status.truecolors[rama & 0x0f];
  PUTPIXEL_BPP3_4(bptr, col);
  PUTPIXEL_BPP3_4(bptr2, col);
  col = status.truecolors[ramb >> 4];
  PUTPIXEL_BPP3_4(bptr, col);
  PUTPIXEL_BPP3_4(bptr2, col);
  col = status.truecolors[ramb & 0x0f];
  PUTPIXEL_BPP3_4(bptr, col);
  PUTPIXEL_BPP3_4(bptr2, col);
}

static void gpl_mode2_bpp4(tt_u8 *dptr, char *bptr) /* bitmap 16 */
{
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  col = status.truecolors[rama >> 4];
  PUTPIXEL_BPP4_4(bptr, col);
  PUTPIXEL_BPP4_4(bptr2, col);
  col = status.truecolors[rama & 0x0f];
  PUTPIXEL_BPP4_4(bptr, col);
  PUTPIXEL_BPP4_4(bptr2, col);
  col = status.truecolors[ramb >> 4];
  PUTPIXEL_BPP4_4(bptr, col);
  PUTPIXEL_BPP4_4(bptr2, col);
  col = status.truecolors[ramb & 0x0f];
  PUTPIXEL_BPP4_4(bptr, col);
  PUTPIXEL_BPP4_4(bptr2, col);
}

static void gpl_mode3_bpp1(tt_u8 *dptr, char *bptr) /* 80 cols */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned char col;
  unsigned char c0 = pixels[0];
  unsigned char c1 = pixels[1];

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? c1 : c0;
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    rama <<= 1;
  }
  for(i = 0; i < 8; i++) {
    col = ramb & 0x80 ? c1 : c0;
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    ramb <<= 1;
  }
}

static void gpl_mode3_bpp2(tt_u8 *dptr, char *bptr) /* 80 cols */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned short col;
  unsigned short c0 = status.truecolors[0];
  unsigned short c1 = status.truecolors[1];

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? c1 : c0;
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    rama <<= 1;
  }
  for(i = 0; i < 8; i++) {
    col = ramb & 0x80 ? c1 : c0;
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    ramb <<= 1;
  }
}

static void gpl_mode3_bpp3(tt_u8 *dptr, char *bptr) /* 80 cols */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;
  unsigned long c0 = status.truecolors[0];
  unsigned long c1 = status.truecolors[1];

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? c1 : c0;
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    rama <<= 1;
  }
  for(i = 0; i < 8; i++) {
    col = ramb & 0x80 ? c1 : c0;
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    ramb <<= 1;
  }
}

static void gpl_mode3_bpp4(tt_u8 *dptr, char *bptr) /* 80 cols */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;
  unsigned long c0 = status.truecolors[0];
  unsigned long c1 = status.truecolors[1];

  for(i = 0; i < 8; i++) {
    col = rama & 0x80 ? c1 : c0;
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    rama <<= 1;
  }
  for(i = 0; i < 8; i++) {
    col = ramb & 0x80 ? c1 : c0;
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    ramb <<= 1;
  }
}

static void gpl_mode4_bpp1(tt_u8 *dptr, char *bptr) /* bitmap 4 special */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned char col;

  for(i = 0; i < 4; i++) {
    col = pixels[((rama & 0x80) >> 7) + ((rama & 0x40) >> 5)];
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    PUTPIXEL_BPP1(bptr2, col);
    rama <<= 2;
  }
  for(i = 0; i < 4; i++) {
    col = pixels[((ramb & 0x80) >> 7) + ((ramb & 0x40) >> 5)];
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr, col);
    PUTPIXEL_BPP1(bptr2, col);
    PUTPIXEL_BPP1(bptr2, col);
    ramb <<= 2;
  }
}

static void gpl_mode4_bpp2(tt_u8 *dptr, char *bptr) /* bitmap 4 special */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned short col;

  for(i = 0; i < 4; i++) {
    col = status.truecolors[((rama & 0x80) >> 7) + ((rama & 0x40) >> 5)];
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    PUTPIXEL_BPP2(bptr2, col);
    rama <<= 2;
  }
  for(i = 0; i < 4; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((ramb & 0x40) >> 5)];
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr, col);
    PUTPIXEL_BPP2(bptr2, col);
    PUTPIXEL_BPP2(bptr2, col);
    ramb <<= 2;
  }
}

static void gpl_mode4_bpp3(tt_u8 *dptr, char *bptr) /* bitmap 4 special */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  for(i = 0; i < 4; i++) {
    col = status.truecolors[((rama & 0x80) >> 7) + ((rama & 0x40) >> 5)];
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    PUTPIXEL_BPP3(bptr2, col);
    rama <<= 2;
  }
  for(i = 0; i < 4; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((ramb & 0x40) >> 5)];
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr, col);
    PUTPIXEL_BPP3(bptr2, col);
    PUTPIXEL_BPP3(bptr2, col);
    ramb <<= 2;
  }
}

static void gpl_mode4_bpp4(tt_u8 *dptr, char *bptr) /* bitmap 4 special */
{
  int i;
  char *bptr2 = bptr + img->bytes_per_line;
  tt_u8 rama = *(dptr + 0x2000);
  tt_u8 ramb = *dptr;
  unsigned long col;

  for(i = 0; i < 4; i++) {
    col = status.truecolors[((rama & 0x80) >> 7) + ((rama & 0x40) >> 5)];
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    PUTPIXEL_BPP4(bptr2, col);
    rama <<= 2;
  }
  for(i = 0; i < 4; i++) {
    col = status.truecolors[((ramb & 0x80) >> 7) + ((ramb & 0x40) >> 5)];
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr, col);
    PUTPIXEL_BPP4(bptr2, col);
    PUTPIXEL_BPP4(bptr2, col);
    ramb <<= 2;
  }
}

__inline__ void quick_decode_gpl(int gpl)
{
  updategpl[(gpl / 320) * 40 + gpl % 40] = 1;
}

static void video_mode_bpp1(void)
{
  switch (status.regE7DC) {
  case 0x00 :
    gpl_func = gpl_mode0_bpp1;
    break;
  case 0x21 :
    gpl_func = gpl_mode1_bpp1;
    break;
  case 0x7b :
    gpl_func = gpl_mode2_bpp1;
    break;
  case 0x2a :
    gpl_func = gpl_mode3_bpp1;
    break;
  case 0x41 :
    gpl_func = gpl_mode4_bpp1;
    break;
  }
}

static void video_mode_bpp2(void)
{
  switch (status.regE7DC) {
  case 0x00 :
    gpl_func = gpl_mode0_bpp2;
    break;
  case 0x21 :
    gpl_func = gpl_mode1_bpp2;
    break;
  case 0x7b :
    gpl_func = gpl_mode2_bpp2;
    break;
  case 0x2a :
    gpl_func = gpl_mode3_bpp2;
    break;
  case 0x41 :
    gpl_func = gpl_mode4_bpp2;
    break;
  }
}

static void video_mode_bpp3(void)
{
  switch (status.regE7DC) {
  case 0x00 :
    gpl_func = gpl_mode0_bpp3;
    break;
  case 0x21 :
    gpl_func = gpl_mode1_bpp3;
    break;
  case 0x7b :
    gpl_func = gpl_mode2_bpp3;
    break;
  case 0x2a :
    gpl_func = gpl_mode3_bpp3;
    break;
  case 0x41 :
    gpl_func = gpl_mode4_bpp3;
    break;
  }
}

static void video_mode_bpp4(void)
{
  switch (status.regE7DC) {
  case 0x00 :
    gpl_func = gpl_mode0_bpp4;
    break;
  case 0x21 :
    gpl_func = gpl_mode1_bpp4;
    break;
  case 0x7b :
    gpl_func = gpl_mode2_bpp4;
    break;
  case 0x2a :
    gpl_func = gpl_mode3_bpp4;
    break;
  case 0x41 :
    gpl_func = gpl_mode4_bpp4;
    break;
  }
}

void gpl_fast_refresh(void)
{
  int x, y, x0, x1, j;
  char *ptr, *bptr0, *bptr2;
  tt_u8 *dptr;
  char *bptr;
  tt_u8 *dptr2;
 
  (*video_mode_func)();

  ptr = updategpl;
  dptr = ramdata + status.vidpage * 0x4000;
  x0 = visible_cols * 16;
  x1 = x0 + 640;
  x = x0;
  y = visible_lines * 2;
  bptr = bptr0 = img->data + x * bytes_per_pixel + img->bytes_per_line * y;
  do {
    if (*ptr) {
      dptr2 = dptr;
      bptr2 = bptr;
      for (j = 0; j < 8; j++) {
	(*gpl_func)(dptr2, bptr2);
	dptr2 += 40;
	bptr2 += img->bytes_per_line * 2;
      }
      *ptr = 0;
      display_update_gpl(x, y);
    }
    x += 16;
    bptr += bytes_per_pixel*16;
    dptr++;
    if (x == x1) {
      dptr += 40*7;
      bptr0 += img->bytes_per_line * 2 * 8;
      bptr = bptr0;
      x = x0;
      y += 16;
    }
  } while (++ptr < updategpl + 1000);

  /*
  display_flush();
  */
}

static void video_screen_refresh(void)
{
  int i = 40, j = 200;
  char *bptr0;
  tt_u8 *dptr;
  char *bptr;
 
  (*video_mode_func)();

  dptr = ramdata + status.vidpage * 0x4000;
  bptr = bptr0 = img->data + visible_cols*16*bytes_per_pixel + img->bytes_per_line*visible_lines*2;
  do {
    (*gpl_func)(dptr, bptr);
    bptr += bytes_per_pixel*16;
    dptr++;
    if (--i == 0) {
      bptr0 += img->bytes_per_line * 2;
      bptr = bptr0;
      i = 40;
      j--;
    }
  } while (j > 0);

  memset(updategpl, 0, 1000);
}

static void video_border_refresh(void)
{
  char *bptr, *bptr0;
  int x,y;

  for (y = 0, bptr = img->data; y < visible_lines; ++y, bptr += img->bytes_per_line*2)
    for (x = 0, bptr0 = bptr; x < 40 + visible_cols*2; x++, bptr0 += bytes_per_pixel*16) {
      (*gpl_border_func)(bptr0);
      (*gpl_border_func)(bptr0 + img->bytes_per_line*(400 + visible_lines*2));
    }

  for (x = 0, bptr = img->data + img->bytes_per_line*visible_lines*2; x < visible_cols; x++, bptr += bytes_per_pixel*16)
    for (y = 0, bptr0 = bptr; y < 200; y++, bptr0 += img->bytes_per_line*2) {
      (*gpl_border_func)(bptr0);
      (*gpl_border_func)(bptr0 + bytes_per_pixel*(640 + visible_cols*16));
    }
}

void video_refresh(void)
{
  video_border_refresh();
  video_screen_refresh();
}

tt_u8 giveE7E7(void)
{
  tt_u8 r;
  int col = cyclecnt % 64;

  if (cyclecnt >= 56*64 && cyclecnt < 256*64)
    r = 0x80;
  else {
    r = 0x00;
  }
  
  if (col >= 12 && col < 52)
    r |= 0x20;

  return r;
}

void video_do_cycle_fast(int cpucycles)
{
  static palette_latched = 0;
  static tt_u8 oldreg = 0xff;

  if (cpucycles > 0) {
    cyclecnt += cpucycles;

#if defined SOUND || defined EXACT_TIMER
    if (status.timer_enabled) {
      timercnt -= cpucycles;
      if (timercnt <= 0) {
	timercnt += 8;
	if (--status.timer_val == 0)
	  timer_timeout();
      }
    }
#endif
#ifdef SOUND
    if (audio) {
      soundcnt -= cpucycles;
      if (soundcnt <= 0) {
	soundcnt += 91;
	audio_put_byte_and_sync(status.mute ? 0x00 : status.piabprbout & status.piabprbmask & 0x3f);
      }
    }
#endif

    if (cyclecnt >= 256*64)
      if (!palette_latched) {
	palette_latched = 1;
	if (need_screen_refresh)
	  video_screen_refresh();
	else
	  gpl_fast_refresh();
	if (need_border_refresh && oldreg == status.regE7DD)
	  need_border_refresh = 0;
	if (need_border_refresh)
	  video_border_refresh();
	if (need_border_refresh || need_screen_refresh) {
	  display_update();
	  need_border_refresh = 0;
	  oldreg = status.regE7DD;
	  need_screen_refresh = 0;
	} 
	if (need_palette_refresh) {
	  set_pseudocolors();
	  need_palette_refresh = 0;
	}
      }

    if (cyclecnt >= 312*64) {
      cyclecnt -= 312*64;
      
#if !defined EXACT_TIMER && !defined SOUND
      if (timer_enabled) {
	tt_s32 newtimer = timer_val;
	
	newtimer -= 312*64/8;
	if (newtimer <= 0) {
	  status.timer_val = 0;
	  timer_timeout();
      } else
	status.timer_val = status.newtimer;
      }
#endif

      frames++;
      display_handle_events();
      palette_latched = 0;
    }
  }
}

void video_do_cycle(int cpucycles)
{
  static tt_u8 *dptr, *rdptr;
  static char *bptr;
  static int cyclecnt_next = 0;
  static int val;
  static int line, bline, col;
  static tt_u8 oldreg;

  if (cyclecnt != cyclecnt_next) { /* invalid ptrs, wait next frame */
    while (cpucycles > 0 && cyclecnt != 0) {
      cpucycles--;
      cyclecnt++;
      /*
      if (status.timer_enabled) {
	timercnt--;
	if (timercnt <= 0) {
	  timercnt = 8;
	  if (--status.timer_val == 0)
	    timer_timeout();
	}
      }
      */
      if (cyclecnt == 312*64) {
	cyclecnt = 0;
	frames++;
      }
    }
    cyclecnt_next = 0;
  }

  if (status.timer_enabled) {
    timercnt -= cpucycles;
    if (timercnt <= 0) {
      timercnt += 8;
      if (--status.timer_val == 0)
	timer_timeout();
    }
  }
#ifdef SOUND
  if (audio) {
    soundcnt -= cpucycles;
    if (soundcnt <= 0) {
      soundcnt += 91;
      audio_put_byte_and_sync(status.mute ? 0x00 : status.piabprbout & status.piabprbmask & 0x3f);
    }
  }
#endif

  while (cpucycles > 0) {

    if (cyclecnt == 0) { /* initialize ptrs */
      dptr = ramdata; 
      (*video_mode_func)();
      rdptr = dptr + status.vidpage * 0x4000;
      oldreg = status.regE7DC;
      bptr = img->data;
      if (visible_lines == 56)
	if (visible_cols == 12)
	  val = VISIBLE_UP;
	else
	  val = INVISIBLE_UP_LEFT;
      else
	val = INVISIBLE_UP;
      line = 0;
      bline = 0;
      col = 0;
    }

    do {
      switch (val) {
      case INVISIBLE_UP :
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  if (line == 56)
	    if (visible_cols == 12)
	      val = VISIBLE_LEFT;
	    else
	      val = INVISIBLE_LEFT;
	  else if (line == 56 - visible_lines)
	    if (visible_cols == 12)
	      val = VISIBLE_UP;
	    else
	      val = INVISIBLE_UP_LEFT;
	}
	break;
      case INVISIBLE_UP_LEFT :
	cyclecnt++;
	col++;
	if (col == 12 - visible_cols)
	  val = VISIBLE_UP;
	break;
      case VISIBLE_UP :
	(*gpl_border_func)(bptr);
	bptr += bytes_per_pixel*16;
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 56)
	    val = VISIBLE_LEFT;
	  else
	    val = VISIBLE_UP;
	} else if (col == 64 - 12 + visible_cols)
	  val = INVISIBLE_UP_RIGHT;
	break;
      case INVISIBLE_UP_RIGHT :
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 56)
	    val = INVISIBLE_LEFT;
	  else
	    val = INVISIBLE_UP_LEFT;
	}
	break;
      case INVISIBLE_LEFT :
	cyclecnt++;
	col++;
	if (col == 12)
	  val = VISIBLE_CENTER;
	else if (col == 12 - visible_cols)
	  val = VISIBLE_LEFT;
	break;
      case VISIBLE_LEFT :
	(*gpl_border_func)(bptr);
	bptr += bytes_per_pixel*16;
	cyclecnt++;
	col++;
	if (col == 12)
	  val = VISIBLE_CENTER;
	break;
      case VISIBLE_CENTER :
	if (oldreg != status.regE7DC) {
	  (*video_mode_func)();
	  oldreg = status.regE7DC;
	  rdptr = dptr + status.vidpage * 0x4000;
	}
	(*gpl_func)(rdptr, bptr);
	dptr++;
	rdptr++;
	bptr += bytes_per_pixel*16;
	cyclecnt++;
	col++;
	if (col == 40+12)
	  if (visible_cols == 0)
	    val = INVISIBLE_RIGHT;
	  else
	    val = VISIBLE_RIGHT;
	break;
      case VISIBLE_RIGHT :
	(*gpl_border_func)(bptr);
	bptr += bytes_per_pixel*16;
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 200+56)
	    if (visible_lines == 0)
	      val = INVISIBLE_DOWN;
	    else
	      val = VISIBLE_DOWN;
	  else
	    val = VISIBLE_LEFT;
	} else if (col == 64 - 12 + visible_cols)
	  val = INVISIBLE_RIGHT;
	break;
      case INVISIBLE_RIGHT :
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 200+56)
	    if (visible_lines == 0)
	      val = INVISIBLE_DOWN;
	    else
	      val = INVISIBLE_DOWN_LEFT;
	  else
	    val = INVISIBLE_LEFT;
	}
	break;
      case INVISIBLE_DOWN_LEFT :
	cyclecnt++;
	col++;
	if (col == 12 - visible_cols)
	  val = VISIBLE_DOWN;
	break;
      case VISIBLE_DOWN :
	(*gpl_border_func)(bptr);
	bptr += bytes_per_pixel*16;
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 200+56 + visible_lines)
	    val = INVISIBLE_DOWN;
	  else
	    val = VISIBLE_DOWN;
	} else if (col == 64 - 12 + visible_cols)
	  val = INVISIBLE_DOWN_RIGHT;
	break;
      case INVISIBLE_DOWN_RIGHT :
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	  bline += 2;
	  bptr = img->data + img->bytes_per_line * bline;
	  if (line == 200+56 + visible_lines)
	    val = INVISIBLE_DOWN;
	  else
	    val = INVISIBLE_DOWN_LEFT;
	}
	break;
      case INVISIBLE_DOWN :
	cyclecnt++;
	col++;
	if (col == 64) {
	  col = 0;
	  line++;
	}
	break;
      }
      /*
      if (status.timer_enabled) {
	timercnt--;
	if (timercnt <= 0) {
	  timercnt = 8;
	  if (--status.timer_val == 0)
	    timer_timeout();
	}
      }
      */
    } while (--cpucycles > 0 && line < 200+56*2);

    if (line == 200+56*2) {
      cyclecnt = 0;
      frames++;
      display_update();
      display_handle_events();
    }
    cyclecnt_next = cyclecnt;
  }
}

void video_dumpstate(void)
{
  printf("Video cycle: %d  Frame: %ld\n", cyclecnt, frames);
}

void video_init(void)
{
  switch (bytes_per_pixel) {
  case 1 :
    video_mode_func = video_mode_bpp1;
    gpl_border_func = gpl_border_bpp1;
    break;
  case 2 :
    video_mode_func = video_mode_bpp2;
    gpl_border_func = gpl_border_bpp2;
    break;
  case 3 :
    video_mode_func = video_mode_bpp3;
    gpl_border_func = gpl_border_bpp3;
    break;
  case 4 :
    video_mode_func = video_mode_bpp4;
    gpl_border_func = gpl_border_bpp4;
    break;
  }
  status.regE7DC = 0;
  (*video_mode_func)();
}






